@@ -34,14 +34,22 @@ static inline unsigned long arch_ffz(unsigned long bitmask)
 	return ia64_popcnt(bitmask & (~bitmask - 1));
 }
 
-static inline unsigned long get_cpu_clock(void)
+static inline unsigned long long get_cpu_clock(void)
 {
-	unsigned long ret;
+	unsigned long long ret;
 
 	__asm__ __volatile__("mov %0=ar.itc" : "=r" (ret) : : "memory");
 	return ret;
 }
 
+#define ARCH_HAVE_INIT
+extern int tsc_reliable;
+static inline int arch_init(char *envp[])
+{
+	tsc_reliable = 1;
+	return 0;
+}
+
 #define ARCH_HAVE_FFZ
 #define ARCH_HAVE_CPU_CLOCK
 
