@@ -2,7 +2,7 @@
 #include "../fio.h"
 
 struct gf_options {
-	struct thread_data *td;
+	void *pad;
 	char *gf_vol;
 	char *gf_brick;
 };
@@ -19,3 +19,4 @@ extern void fio_gf_cleanup(struct thread_data *td);
 extern int fio_gf_get_file_size(struct thread_data *td, struct fio_file *f);
 extern int fio_gf_open_file(struct thread_data *td, struct fio_file *f);
 extern int fio_gf_close_file(struct thread_data *td, struct fio_file *f);
+extern int fio_gf_unlink_file(struct thread_data *td, struct fio_file *f);
