@@ -2,6 +2,7 @@
 #define FIO_MUTEX_H
 
 #include <pthread.h>
+#include "lib/types.h"
 
 #define FIO_MUTEX_MAGIC		0x4d555445U
 #define FIO_RWLOCK_MAGIC	0x52574c4fU
@@ -30,7 +31,7 @@ extern void __fio_mutex_remove(struct fio_mutex *);
 extern void fio_mutex_remove(struct fio_mutex *);
 extern void fio_mutex_up(struct fio_mutex *);
 extern void fio_mutex_down(struct fio_mutex *);
-extern int fio_mutex_down_trylock(struct fio_mutex *);
+extern bool fio_mutex_down_trylock(struct fio_mutex *);
 extern int fio_mutex_down_timeout(struct fio_mutex *, unsigned int);
 
 extern void fio_rwlock_read(struct fio_rwlock *);
